# TimetableColspan - An implementation of Timetables that includes
#                    the ability for an event to span days, a
#                    Zope Page Template (ZPT) implementation.
#
# Copyright (C) 2002 Robert A. Klahn (rklahn@acm.org)
#               2000-2002 Juan David Ibez Palomar <jdavid@nuxeo.com>
#               2000-2001 Universitat Jaume I (http://www.uji.es)
#               2002      Nuxeo (http://www.nuxeo.com)
#
# This program is a modification of the entirety of "Timetable Widget",
# a work by Juan David Ibez Palomar, Universitat Jaume I, and Nuxeo.
# This modification began (GNU License condition 2a) on March 14th, 2002,
# each source file has been modified, and potentially renamed. The entire
# modification is licensed under the GNU General Public License, as stated
# below (GNU License condition 2b). The modified program is not interactive
# in nature (GNU License condition 2c).
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

"""TimetableColspan"""
__version__ = "$Id: __init__.py,v 1.6 2002/08/26 17:53:32 rklahn Exp $"

# declare this module public
from AccessControl import ModuleSecurityInfo
ModuleSecurityInfo('Products').declarePublic('TimetableColspan')
__roles__=None
security=ModuleSecurityInfo()

# The template with the macro
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
timetable = PageTemplateFile('timetable', globals())
timetable.__roles__=None

import copy

class Time:
    """
    This class represents a positive amount of time with a precision of
    minutes. Internally time is represented as a pair of values, hours
    and minutes.
    """

    def __init__(self, time):
        """
        The constructor takes an string with the format "hh:mm", it
        parses the string and internally stores the hours and minutes
        as integers.
        """

        x = time.split(':')

        if len(x) != 2:
            raise ValueError, time

        self.hh, self.mm = int(x[0]), int(x[1])

    def __cmp__(self, other):
        """Compares two Time instances."""

        sh, sm = self.hh, self.mm
        oh, om = other.hh, other.mm
        if sh > oh:
            return 1
        elif sh == oh:
            if sm > om:
                return 1
            elif sm == om:
                return 0
            else:
                return -1
        else:
            return -1

    def __sub__(self, other):
        """
        Calculates the time between two Time objects, the result
        is a real number, the units is hours.
        """

        sh, sm = self.hh, self.mm
        oh, om = other.hh, other.mm
        return (sh - oh) + (sm - om)/60.0

    def __add__(self, other):
        """ Calculates the sum of two Time objects. The result is
            a Time object
        """
        sh, sm = self.hh, self.mm
        oh, om = other.hh, other.mm
        totalmin = ((self.hh+other.hh)*60)+self.mm+other.mm
        totalhours = totalmin/60
        totalmin = totalmin - (totalhours*60)
        return Time('%02d:%02d' % (totalhours, totalmin))

    def str24(self):
        """ Textual representation of a Time object, in 24 hour format.
        """

        return "%02d:%02d" % (self.hh,self.mm)

    def __str__(self):
        """
        Textual representation of a Time object, it's in 12 hour format.
        """

        phour = self.hh
        ampm = "AM"
        if phour == 0:
            phour = 12
        elif phour == 12:
            ampm = "PM"
        elif phour > 12:
            phour -= 12
            ampm = "PM"
        
        return "%02d:%02d %s" % (phour, self.mm, ampm)

class SCell:
    """ A Class which represents a cell in the schedule.

        A SCell contains a day, and a (start) time, and a title.
    """
    def __init__(self, time, day):
        self.day = day
        self.time = time
        self.title = None

    def __str__(self):
        return "day: "+str(self.day)+" time: "+str(self.time)+" title: "+str(self.title)

    def set(self,title):
        self.title = title

    def rettitle(self):
        return self.title

class Schedule:
    """ Contains a two dimensional list of cells. The two dimensions are times and days.
        Initialy all cells are empty. Only non-empty cells can be inserted via the
        methods of this class.
    """
    def __init__(self, times, days):
        self.times = times
        self.days = days
        self.cell = []
        for time in times:
            crow = []
            for day in days:
                crow.append(SCell(time,day))
            self.cell.append(crow)

    def __colapse__(self, timeindex, dayindex):
        """ Given an index into the self.cell list o' lists, calculate the
            rowspan and colspan for a table cell at that location. Mark
            any eligible cells as collapsed, by replacing their cell with None
        """
        rowspan=1
        colspan=1
        event=self.cell[timeindex][dayindex].rettitle()
        while dayindex+colspan < len(self.cell[timeindex]):
            # Never match to a cell that has already been colapsed.
            if self.cell[timeindex][dayindex+colspan] is None: 
                break
            if event == self.cell[timeindex][dayindex+colspan].rettitle():
                self.cell[timeindex][dayindex+colspan] = None
                colspan+=1
            else:
                break
        while timeindex+rowspan < len(self.cell):
            # This loop checks to see if each column in this row has
            # the same event, to the extent of the columns in the first
            # row
            breakwhile = (1==0)
            for curcol in range(dayindex,dayindex+colspan):
                if event != self.cell[timeindex+rowspan][curcol].rettitle():
                    breakwhile=(1==1) # row does NOT qualify
                    break # break from for
            if breakwhile:
                break # break from while
            # got this far, then the row qualifies for colapsing into the cell
            for curcol in range(dayindex,dayindex+colspan):
                self.cell[timeindex+rowspan][curcol]=None
            rowspan+=1
            
        return rowspan,colspan,event

    def __insert__(self,itime):
        """ Insert a time into self.times, and make a corresponding insert into
            self.cell
        """
        index = len(self.times)
        for cmp_times in self.times:
            if itime < cmp_times:
                # the position in the list where we want to insert
                index = self.times.index(cmp_times)
                break
        self.times.insert(index,itime)
        if index == 0:
            rowcopy = [ None for x in self.days ]
        else:
            rowcopy = copy.deepcopy(self.cell[index-1])
        self.cell.insert(index,rowcopy)

    def set(self,day_start,day_end,time_start,time_end,item):
        """ Insert a new event into the timetable. No event can overlap, so
            if one does, raise an error.
        """
        # if time_start or time_end is not in self.times, then we need to insert it (either)
        # into self.times and a row needs to be inserted (in the right place, of course)
        # into self.cell. 
        if time_start not in self.times:
            self.__insert__(time_start)
                
        if time_end not in self.times:
            self.__insert__(time_end)

        # And now to insert the event into the cells
        try:
            timeindexend = self.times.index(time_end)
        except:
            timeindexend = len(self.times)
        for timeindex in range(self.times.index(time_start),timeindexend):
            for dayindex in range(self.days.index(day_start),self.days.index(day_end)+1):
                if self.cell[timeindex][dayindex].rettitle() is not None:
                    raisemess= "day_start: "+str(day_start)+" day_end: "+str(day_end)
                    raisemess+=" time_start: "+str(time_start)+" time_end: "+str(time_end)
                    raisemess+=" title: "+(item)
                    raise 'overlapping timetable entry',raisemess
                else:
                    self.cell[timeindex][dayindex].set(item)
        return 0

    def render(self):
        """ Builds an array representing the data elements to the table
            to be displayed, with rowspan and colspan, then returns
            this object in total.
        """
        self.body = []
        timeindex=0
        for time in self.times[:-1]: # last item is the final end time, it does not appear as a start time
            row = {}
            row['start'] = str(time)
            row['start24'] = time.str24()
            row['end24'] = self.times[timeindex+1].str24()
            row['t_items'] = []
            dayindex=0
            for day in self.days:
                item = {}
                # if it is None, then this is a cell we have colapsed into
                # a rowspan/colspan
                if self.cell[timeindex][dayindex] is not None:
                    item['rowspan'],item['colspan'],item['title']=self.__colapse__(timeindex,
                                                                                   dayindex)
                    item['starttime']=time.str24()
                    item['startday']=day
                    try: # Might be beyond end time
                        item['endtime']=self.times[self.times.index(time)+item['rowspan']].str24()
                    except:
                        item['endtime']=row['end24'];
                    item['endday']=self.days[self.days.index(day)+item['colspan']-1]
                    row['t_items'].append(item)
                dayindex+=1
            timeindex+=1
            self.body.append(row)
        return self

Schedule.__allow_access_to_unprotected_subobjects__ = 1
        
def get_data(data, times, days):
    """ """

    # Parameter validity
    if data is None:
        data = []

    if times is None:
        times = [ Time(x) for x in ['0:00','1:00','2:00','3:00','4:00',
                                    '5:00','6:00','7:00','8:00','9:00',
                                    '10:00','11:00','12:00','13:00','14:00',
                                    '15:00','16:00','17:00','18:00','19:00',
                                    '20:00','21:00','22:00','23:00','24:00'] ]
    else:
        times = [ Time(x) for x in times ]
    times.sort()
    

    if days is None:
        days = [None]

    # Validate data, build DataGrid.
    grid = Schedule(times,days)
    for event in data:
        try: # might not exist, if there is only one column
            day_start = event['day_start']
        except:
            day_start = None
        try:
            day_end = event['day_end']
        except: # default is day_start
            day_end = day_start
        # Required. Permit error to be raised.
        time_start = Time(event['time_start'])
        time_end = Time(event['time_end'])
        title = event['title']
        grid.set(day_start,day_end,time_start,time_end,title)

    return grid.render()
        
security.declarePublic('get_data')
security.apply(globals())


